<?php

declare(strict_types=1);

namespace mirzaev\parser_from_interneturok\models;

// Files of the project
use mirzaev\parser_from_interneturok\models\core,
	mirzaev\parser_from_interneturok\models\enumerations\subject,
	mirzaev\parser_from_interneturok\models\homework;

// Svoboda time
use svoboda\time\statement as svoboda;

// Baza database
use mirzaev\baza\record;

// Browser
use GuzzleHttp\Client as client,
	GuzzleHttp\Psr7\Request as request,
	GuzzleHttp\Promise\PromiseInterface as promise;

// Framework for excel
use avadim\FastExcelWriter\Excel as excel;

// Built-in libraries
use Exception as exception,
	RuntimeException as exception_runtime,
	SplFileObject as spl,
	DOMDocument as dom,
	DOMXPath as xpath,
	DateTime as datetime;

/**
 * InternetUrok
 *
 * @see https://interneturok.ru/
 *
 * @package mirzaev\parser_from_interneturok\models
 *
 * @license http://www.wtfpl.net/ Do What The Fuck You Want To Public License
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
final class interneturok extends core
{
	/**
	 * ROOT
	 *
	 * Method: GET
	 *
	 * @var string ROOT The root URN
	 */
	public const string ROOT = 'https://interneturok.ru';

	/**
	 * AUTHENTICATION
	 *
	 * Method: GET
	 *
	 * @var const string AUTHENTICATION The authentication API URL
	 */
	public const string AUTHENTICATION = 'https://api-gw.interneturok.ru/api/v2/psp/tokens';

	/**
	 * USER
	 *
	 * Method: GET
	 *
	 * @var const string USER The user data API URL
	 */
	public const string USER = 'https://api-gw.interneturok.ru/api/v2/homeschool/current_user';

	/**
	 * SCHEDULES
	 *
	 * Method: GET
	 * Authrorization: Bearer
	 *
	 * @var const string SCHEDULES The user schedules API URL
	 */
	public const string SCHEDULES = 'https://api-gw.interneturok.ru/api/v2/schedules/study-years?filter[status][eq]=current';

	/**
	 * JOURNAL
	 *
	 * Method: GET
	 * Format: "https://api-ege.interneturok.ru/api/v2/journal/student?grade=$grade&quarter=$quarter&year_id=$year&user_id=$identifier&token=$this->token"
	 * Authrorization: Bearer
	 *
	 * @var const string JOURNAL The user journal API URL
	 */
	public const string JOURNAL = 'https://api-ege.interneturok.ru/api/v2/journal/student';

	/**
	 * TIME
	 *
	 * Method: GET
	 *
	 * @var const string TIME The moscow time API URL
	 */
	public const string TIME = 'https://api-gw.interneturok.ru/api/v2/settings/moscow-time';

	/**
	 * LESSONS
	 *
	 * Method: GET
	 * Format: "https://api-gw.interneturok.ru/api/v2/homeschool/lessons/$lesson?items=true"
	 * Authrorization: Bearer
	 *
	 * @var const string LESSONS The lesson data API URL
	 */
	public const string LESSONS = 'https://api-gw.interneturok.ru/api/v2/homeschool/lessons';

	/**
	 * TESTS
	 *
	 * Method: GET
	 * Format: "https://api-gw.interneturok.ru/api/v2/tests/users/current/tests/$test/result"
	 * Authrorization: Bearer
	 *
	 * @var const string TESTS The lesson data API URL
	 */
	public const string TESTS = 'https://api-gw.interneturok.ru/api/v2/tests/users/current/tests';

	/**
	 * ATTACHMENTS
	 *
	 * Method: GET
	 * Format: "https://api-gw.interneturok.ru/api/v2/homeschool/attachments?attachable_id=$attachment&attachable_type=Result::Homework"
	 * Authrorization: Bearer
	 *
	 * @var const string ATTACHMENTS The attachments data API URL
	 */
	public const string ATTACHMENTS = 'https://api-gw.interneturok.ru/api/v2/homeschool/attachments';

	/**
	 * Browser
	 *
	 * @var client $browser The guzzle client
	 */
	private client $browser;

	/**
	 * Token
	 *
	 * The account authorization bearer token
	 *
	 * @var string $token The account authorization bearer token
	 */
	private string $token;

	/**
	 * Constructor
	 */
	public function __construct()
	{
		// Initializint the guzzle client
		$this->browser = new client(['cookies' => true]);
	}

	/**
	 * Parse
	 *
	 * @param subject $subject The subject
	 * @param int $grade The grade
	 * @param int $waiting Interval for processing requests (seconds)
	 *
	 * @return array|false Downloaded homework files
	 */
	public function parse(subject $subject, int $grade, int $waiting = 3): array|false
	{
		// Initializing accounts
		$accounts = static::accounts();

		if (!empty($accounts)) {
			// Initialized accounts

			/* if (shuffle($accounts)) { */
			// Shuffled accounts

			foreach ($accounts as $account) {
				// Iterating over accounts

				// Waiting for processing the request
				sleep($waiting);

				if ($this->authentication($account)->wait()) {
					// Authenticated the account

					// Writing into the output buffer
					echo 'Аутентифицирован: ' . $account['mail'] . "\n";

					// Waiting for processing the request
					sleep($waiting);

					// Initializing the user journal
					$journal = $this->journal(grade: $grade)->wait();

					if (!empty($journal)) {
						// Initialized the user journal

						// Writing into the output buffer
						echo "Инициализирован журнал\n";

						// Initializing the actual date
						$date = new datetime();

						// Initializing the homeworks database
						$model = new homework();

						// Initializing the schedule subject events
						$schedule = $journal?->schedule_events;

						foreach ($schedule as $_subject) {
							// Iterating over scheduled subject events

							if ($_subject->subject?->name === $subject->value) {
								// Found the target suject

								foreach ($_subject->events as $event) {
									// Iterating over the target subject events

									// Normalizing the event variable (events can be more than one)
									$event = $event[0] ?? null;

									if (!empty($event)) {
										// Initialized the target subject event (the homework exists)

										if (new datetime($event->date) <= $date && new datetime($event->date)->modify('+6 days') >= $date) {
											// Found the current event

											if ($event->subject?->name === $subject->value && $event->subject->grade === $grade) {
												// Found the target subject

												// Writing into the output buffer
												echo "Найден школьный предмет: $subject->value для $grade класса (с " . new datetime($event->date)->format('d.m') . ' по ' . new datetime($event->date)->modify('+6 days')->format('d.m') . ")\n";

												// Initializing the unblock time
												$unblock = svoboda::timestamp() - 31536000;

												foreach ($event->homeworks as $homework) {
													// Iterating over scheduled event homeworks

													if ($homework->status === 'checked' && $homework->mark === 5) {
														// Homework checked and completed for a grade of 5

														// Checking for the block record
														$blocked = $model->database->read(
															filter: fn(record $record) => $record->identifier === $homework->item_id && $record->created > $unblock,
															amount: 1
														)[0] ?? null;

														if ($blocked instanceof record) {
															// The account homework has been downloaded for 1 year

															continue 3;
														} else {
															// The account homework has not been downloaded for 1 year

															// Writing into the output buffer
															echo "Найдено домашнее задание: $homework->item_id с оценкой $homework->mark\n";

															// Waiting for processing the request
															sleep($waiting);

															// Initializing the homework files
															$files = $this->homework(lesson: $event->lesson->id, homework: $homework->item_id)->wait();

															if (!empty($files)) {
																// Initialized the homework files

																// Writing into the output buffer
																echo "Получено домашнее задание\n";

																// Blocking the homework downloading for 1 year
																$blocked = $model->create(identifier: $homework->item_id);

																if ($blocked !== false) {
																	// The homework was blocked for 1 year

																// Writing into the output buffer
																echo "Заблокировано на 1 год домашнее задание\n";

																	// Exit (success)
																	return $files;
																}
															}
														}
													}
												}
											}

											// Stopping processing events and starting processing the next account
											break 2;
										}
									}
								}

								// Stopping processing events and starting processing the next account
								break;
							}
						}
					} else {
						// Not initialized the user journal

						// Writing into the output buffer
						echo "Аккаунт не подходит под требования\n";

						// Proceed to processing the next account
						continue;
					}
				} else {
					// Not authenticated the account

					// Writing into the output buffer
					echo "Не аутентифицирован\n";

					// Proceed to processing the next account
					continue;
				}
			}
			/* } */
		}

		// Exit (fail)
		return false;
	}

	/**
	 * Authentication
	 *
	 * Authenticate on the InternetUrok site and save cookies in guzzle http client
	 *
	 * @param array $account The account ['mail' => '', 'password' => '']
	 *
	 * @return promise Is the account authenticated?
	 */
	private function authentication(array $account): promise
	{
		// Initializing the request to the root page
		$request = new request('GET', static::ROOT);

		// Authenticating and exit
		return $this->browser->sendAsync($request)
			->then(function ($response) use ($account) {
				// Sended the request and received the response

				// Initializing the DOM implementator
				$dom = new dom;

				// Parsing the HTML-document
				@$dom->loadHTML((string) $response->getBody());

				// Initializing the XPATH implementator
				$xpath = new xpath($dom);

				// Searching for the profile button
				$entries = $xpath->query('//div[@class="header-userbox show"]/div[@class="header-profile btn btn-header-profile btn-white-border"]');

				if (!empty($entries[0])) {
					// Authenticated (found the profile button)

					// Initializing the profile button
					$profile = $entries[0];

					// Searching for the account mail
					$entries = $xpath->query('//div[@class="header-profile-menu"]/a/div[@class="header-profile-userinfo-text"]/div[@class="header-profile-email js-userprofile-email"]');

					if (!empty($entries[0])) {
						// Found the account mail

						// Initializing the authenticated account mail
						$mail = $entries[0]->textContent;

						if ($account['mail'] === $mail) {
							// Authenticated the target account

							// Exit (success)
							return true;
						} else {
							// Authenticated another account

							// Reinitializing the guzzle client
							$this->browser = new client(['cookies' => true]);

							// Exit (success)
							return false;
						}
					} else {
						// Not found the account mail

						return false;
					}
				} else {
					// Not authenticated (not found the profile button)

					// Exit (success)
					return false;
				}
			})
			->then(
				function (bool $authenticated) use ($account) {
					if ($authenticated) {
						// Authenticated

						// Exit (success)
						return true;
					} else {
						// Not authenticated

						// Initializing the request for authentication
						$request = new request(
							'POST',
							static::AUTHENTICATION,
							[
								'Content-Type' => 'application/x-www-form-urlencoded; charset=UTF-8'
							],
							'login=' . urlencode($account['mail']) . '&password=' . urlencode($account['password'])
						);

						// Sending the request
						return $this->browser->sendAsync($request)
							->then(function ($response) use ($account) {
								// Received the response

								if ($response->getStatusCode() === 200) {
									// Success (received the "OK" status code)

									// Initializing the authentication response body
									$authenticated = json_decode((string) $response->getBody());

									// Initializing the account access token
									$this->token = $authenticated?->data?->psp?->response?->data?->access_token;

									// Exit (success)
									return true;
								} else {
									// Fail (received not the "OK" status code)

									// Exit (fail)
									return false;
								}
							})
							->wait();
					}
				}
			);
	}

	/**
	 * Journal
	 *
	 * Search for the user homeworks journal
	 *
	 * @param int $grade The grade
	 *
	 * @return promise|null The journal data (object)
	 */
	private function journal(int $grade): promise|null
	{
		if (!empty($this->token)) {
			// Initialized the account authorization token

			// Initializing the request to the user data API
			$request = new request(
				'GET',
				static::USER,
				[
					'Authorization' => "Bearer $this->token"
				]
			);

			return $this->browser->sendAsync($request)
				->then(function ($response) use ($grade) {
					// Sended the request and received the response

					// Initializing the user data
					$user = json_decode((string) $response->getBody())?->data?->homeschool?->response?->user;

					if ($user->grade + 1 === $grade) {
						// Matched the account grade with the grade

						// Initialiint the user identifier
						$identifier = $user?->id;

						if (!empty($identifier)) {
							// Initialized the user identifier

							// Initializing the request to the user schedules API
							$request = new request(
								'GET',
								static::SCHEDULES,
								[
									'Authorization' => "Bearer $this->token"
								]
							);

							return $this->browser->sendAsync($request)
								->then(function ($response) use ($grade, $identifier) {
									// Sended the request and received the response

									// Initializing the user schedules
									$schedules = json_decode((string) $response->getBody())?->data?->schedules?->response?->data;

									// Initializing the study years registry
									$years = $schedules?->study_years;

									if (!empty($years)) {
										// Initialized the study years registry

										foreach ($years as $year) {
											// Iterating over study years

											if ($year->status === 'current') {
												// The current study year	

												// Initializing the request to the moscow time API
												$request = new request(
													'GET',
													static::TIME,
													[
														'Authorization' => "Bearer $this->token"
													]
												);

												return $this->browser->sendAsync($request)
													->then(function ($response) use ($grade, $identifier, $year) {
														// Sended the request and received the response

														// Initializing the moscow time
														$time = json_decode((string) $response->getBody())?->data?->time?->response;

														if (!empty($time)) {
															// Initialized the moscow time

															// Implementing the moscow time
															$now = new datetime($time);

															// Declaring the quarter buffer
															$quarter = null;

															foreach ($year->quarters as $value) {
																// Iterating over study year quarters

																// Implementing the quarter start time and end time
																$start = new datetime($value->starts_at);
																$end = new datetime($value->ends_at);

																if ($now > $start && $now < $end) {
																	// Found the current quarter

																	// Writing the current quarter identifier into the quarter buffer
																	$quarter = $value->quarter;

																	// Stopping iterating over study year quarters
																	break;
																}
															}

															// Initializing the request to the user journal API
															$request = new request(
																'GET',
																static::JOURNAL . "?grade=$grade&quarter=$quarter&year_id=$year->id&user_id=$identifier&token=$this->token"
															);

															return $this->browser->sendAsync($request)
																->then(function ($response) {
																	// Sended the request and received the response

																	// Initializing the user journal
																	$journal = json_decode((string) $response->getBody());

																	if (!empty($journal)) {
																		// Initialized the user journal

																		// Exit (success)
																		return $journal;
																	} else {
																		// Not initialized the user journal

																		// Exit (fail)
																		return false;
																	}
																})->wait();
														} else {
															// Not initialized the moscow time

															// Exit (fail)
															return false;
														}
													})->wait();
											}
										}
									} else {
										// Not initialized the study years registry

										// Exit (fail)
										return false;
									}
								})->wait();
						} else {
							// Not initialized the user identifier

							// Exit (fail)
							return false;
						}
					} else {
						// Not matched the account grade with the grade

						// Exit (fail)
						return false;
					}
				});
		}

		// Exit (fail)
		return null;
	}

	/**
	 * Homework
	 *
	 * Search for lesson and homework tetst or attachments, then download and generate excel documents
	 *
	 * @param int $lesson The lesson identifier
	 * @param int $homework The homework identifier
	 *
	 * @return promise|null The homework files (array)
	 */
	private function homework(int $lesson, int $homework): promise|null
	{
		if (!empty($this->token)) {
			// Initialized the account authorization token

			// Initializing the request to the lesson data API
			$request = new request(
				'GET',
				static::LESSONS . "/$lesson?items=true",
				[
					'Authorization' => "Bearer $this->token"
				]
			);

			return $this->browser->sendAsync($request)
				->then(function ($response) use ($homework) {
					// Sended the request and received the response

					// Initializing the lesson
					$lesson = json_decode((string) $response->getBody())?->data?->homeschool?->response;

					foreach ($lesson->items as $section) {
						// Iterating over the lesson sections

						if ($section->id === $homework && $section->item_type === 'homework' && $section->homework_kind === 'homework_base') {
							// Found the base homework section

							// Initializng the test identifier (string)
							$test = $section->homework_attributes?->parsed_content?->evo_test_id ?? null;

							if (!empty($test)) {
								// Initialized the test identifier

								// Initializing the request to the test data API
								$request = new request(
									'GET',
									static::TESTS . "/$test/result",
									[
										'Authorization' => "Bearer $this->token"
									]
								);

								return $this->browser->sendAsync($request)
									->then(function ($response) use ($homework) {
										// Sended the request and received the response

										// Initializing the test result
										$result = json_decode((string) $response->getBody())?->data?->tests?->response?->data;

										if (!empty($result)) {
											// Initialized the test result

											// Initializing paths to the homework storage folder
											$storage = STORAGE . DIRECTORY_SEPARATOR . 'homeworks' . DIRECTORY_SEPARATOR . $homework;

											// Initializing the homework storage folder
											if (!file_exists($storage)) mkdir($storage, 0775, true);

											// Initializing the excel document
											$excel = excel::create(['Домашнее задание']);

											// Initializing the excel document sheet
											$sheet = $excel->sheet();

											// Writing settings of the excel document sheet
											$sheet->setColWidths(['A' => 10, 'B' => 80, 'C' => 50]);

											// Writing title
											$sheet->writeRow(
												['Номер', 'Задание', 'Ответ'],
												[
													'font' => [
														'style' => 'bold'
													],
													'text-align' => 'center',
													'vertical-align' => 'center',
													'border' => 'thin',
													'height' => 24,
												]
											);

											foreach ($result->test?->questions as $question) {
												// Iterating over questions

												foreach ($question->answer_results as $answer) {
													// Iterating over question answers

													if ($answer->is_correct) {
														// The answer is correct
														
														if (!empty($answer->section)) {
															// Several sections for answers (expected)
													
															// Writing questions and answers
															$sheet->writeRow([$question->test_question_position, $question->name, $answer->section->name . ' - ' . $answer->value]);
														} else {
															// One single answer (expected)
													
															// Writing questions and answers
															$sheet->writeRow([$question->test_question_position, $question->name, $answer->value]);
														}
													}
												}
											}

											// Initializing path to the excel document file 
											$path = $storage . DIRECTORY_SEPARATOR . 'homework.xlsx';

											// Writing the excel document file to the storage
											$excel->save($path);

											// Exit (success)
											return [$path];
										} else {
											// Not initialized the test result

											// Exit (fail)
											return false;
										}
									})->wait();
							} else {
								// Not initialized the test identifier

								// Initializing the attachment identifier
								$attachment = $section->result?->id;

								if (!empty($attachment)) {
									// Initialized the attachment identifier

									// Initializing the request to the test data API
									$request = new request(
										'GET',
										static::ATTACHMENTS . "?attachable_id=$attachment&attachable_type=Result::Homework",
										[
											'Authorization' => "Bearer $this->token"
										]
									);

									return $this->browser->sendAsync($request)
										->then(function ($response) use ($homework) {
											// Sended the request and received the response

											// Initializing	attachments
											$attachments = json_decode((string) $response->getBody())?->data?->homeschool?->response?->attachments;

											if (!empty($attachments)) {
												// Initialized attachmetns

												// Initializing paths to the homework attachments storage folder
												$storage = STORAGE . DIRECTORY_SEPARATOR . 'homeworks' . DIRECTORY_SEPARATOR . $homework . DIRECTORY_SEPARATOR . 'attachments';

												// Initializing the homework attachments storage folder
												if (!file_exists($storage)) mkdir($storage, 0775, true);

												// Declaring the buffer of files
												$files = [];

												foreach ($attachments as $attachment) {
													// Iteration over attachments

													if ($attachment->uploader_role === 'student' && $attachment->attachable_type === 'Result::Homework') {
														// Found the homework attachment

														// Initializing path to the downloaded file
														$path = $storage . DIRECTORY_SEPARATOR . $attachment->attach_path;

														// Downloading the file
														$this->browser->request('GET', $attachment->attach, ['sink' => $path]);

														// Writing path to the downloaded file into the buffer of files
														$files[] = $path;
													}
												}

												// Exit (success)
												return $files;
											} else {
												// Not initialized attachments

												// Exit (fail)
												return false;
											}
										})->wait();
								} else {
									// Not initialized the attachment identifier

									// Exit (fail)
									return false;
								}
							}
						}
					}

					// Exit (fail)
					return false;
				},
				function ($response) use ($homework) {
					if ($response->getStatusCode() === 402) {
						// Fail (received the "Payment Required" status code)

						// Writing into the output buffer
						echo "Не удалось скачать домашнее задание $homework потому, что не оплачен тариф\n";

						// Exit (fail)
						return false;
					}
				});
		}

		// Exit (fail)
		return null;
	}


	/**
	 * Accounts
	 *
	 * Read CSV rows and generate account registry
	 *
	 * @return array|false The accounts registry ['mail' => '', 'password' => '']
	 */
	public static function accounts(): array|false
	{
		// Opening the InternetUrok accounts file
		$file = fopen(INTERNETUROK_ACCOUNTS_FILE, "r");

		if (is_resource($file)) {
			// Opened the InternetUrok accounts file

			if (flock($file, LOCK_SH)) {
				// Locked the InternetUrok accounts file

				// Initializing the InternetUrok accounts registry
				$accounts = [];

				while (!empty($row = fgetcsv($file, 5000, " "))) {
					// Readed the CSV row

					// Initializing the account data
					$mail = $row[0];
					$password = $row[1];

					// Writing into the InternetUrok accounts registry
					$accounts[] = ['mail' => $mail, 'password' => $password];
				}

				// Unlocking the InternetUrok accounts file
				flock($file, LOCK_UN);

				// Closing the InternetUrok accounts file
				fclose($file);

				// Exit (success)
				return $accounts;
			}
		}

		// Exit (fail)
		return false;
	}


	/**
	 * Amount
	 *
	 * Calculate the number of rows in the accounts file (the number of accounts is expected)
	 *
	 * @return int Amount of accounts
	 */
	public static function amount(): int
	{
		// Opening the interneturok accounts file
		$file = new spl(INTERNETUROK_ACCOUNTS_FILE, 'r');

		// Move the cursor to the last row
		$file->seek(PHP_INT_MAX);

		// Initializing number of the last row
		$rows = $file->key();

		// Exit (success)
		return $rows;
	}
}
